; Register assignment
;
; ebx   - Top of stack (value, not a pointer)
; esp   - Data stack pointer (data stack is on hardware stack)
; ebp   - Return stack pointer
; esi   - Interpretation pointer, points to the address of the header of the
;         word which will be executed next iteration

; The Forth system can be in two modes: interpretation or compilation
; It will start in interpretation mode, and the current mode will be stored in
; the "mode" variable in the .data section
mode_compile    = 0
mode_interpret  = 1

; Macro to create a word header, of the format documented in the definition.
; See the code for examples of its use.

macro word code,name,flags {
    local this_word
this_word:              ; Define the start of the current header
    dd code             ; 00    Store pointer to code of this word
    dd previous_word    ; 04    Store pointer to header of last defined word
    dd flags            ; 08    Store flags
    db name             ; 0c    Store name string passed as macro parameter
    db 0                ;       Zero-terminate name string
previous_word = this_word   ; This will make the next word link to this one
    align 4             ; Align to a dword - this will speed things up, but
}                       ; isn't strictly necessary


; Allowed flags:

flag_none       = 0             ; A normal word
flag_immediate  = 1             ; An immediate word (will be run compile time)

; The "previous_word" constant (or, actually, a variable inside the assembler)
; always points to the previously defined word of the words defined with the
; "word" macro above. It starts at 0, to denote the end of the built-in word
; list that is generated during assembly.

previous_word   = 0

format ELF                      ; Output an ELF object file for linking
section '.text' executable      ; The code section

; libc exports
public main         ; main()

; libc imports
extrn exit          ; exit(status)
extrn getchar       ; int getchar()
extrn printf        ; printf(format, ...)
extrn putchar       ; putchar(character)
extrn strcmp        ; int strcmp(s1, s2)
extrn strtol        ; long int strtol(str, *endptr, base)
extrn usleep        ; void usleep(unsigned long usec)

; ---------- Forth "inner interpreter" primitives ----------
;
; Here a few basic words and macros are defined.
;
; next      - macro, which will execute the next instruction of a high-level
;             Forth word. Put this on the end of every assembly-level word
;             definition. Making this a macro will make the built-in functions
;             faster and, assuming that the lodsd instruction is used,
;             actually smaller as well.

macro next {
    lodsd           ; pointer to header address = [ip++]
    jmp dword[eax]  ; jump to the code (which is stored in the beginning of
}                   ; the header

; nest      - enter a new high-level word. A call to this should be made in
;             the beginning of each high-level word. The general structure of
;             such a word is:
;
;   call nest
; start_of_highlevel_string:
;   dd address_of_header_of_foo
;   dd address_of_header_of_bar
;   ...
;   dd address_of_header_of_exit
;
;           Note that "nest" is never used explicitly, a call to it will
;           however be compiled by the ":" (define word) word.

nest:
    sub ebp,4           ; Allocate 4 new bytes on the return stack
    mov [ebp],esi       ; Store the current interpretation pointer there
    pop esi             ; Get "start_of_highlevel_string" from the example
    next                ; and put it into esi. Then execute the word there
                        ; (in the example above, "foo")

; exit      - exit a high-level word. Has to be in the end of a high-level
;             definition (so ";' - the "end of word" word - will compile an
;             "exit" word. "exit" can also be used explicitly, like a "return"
;             statement in other languages.

code_exit:
    mov esi,[ebp]       ; Get back our old interpretation pointer
    add ebp,4           ; Free 4 bytes from the return stack
    next                ; Jump to the new location

; execute   - call the word whose header is passed (as an address) on the
;             data stack

code_execute:
    mov eax,ebx         ; Save top of data stack (address to word header)
    pop ebx             ; Remove this stack item
    jmp dword[eax]      ; Jump to the code of the given word

; lit       - push a literal value (loaded from the position right after the
;             lit instruction

code_lit:
    push ebx            ; Save old top of data stack
    mov ebx,[esi]       ; Load its new value from at the interpretation pointer
    add esi,4           ; Advance pointer (so that we won't try to execute the
    next                ; literal value, thit causes trouble)

; nop       - dummy word (used as a placeholder in newly created headers)

code_nop:
    next

; --------- Basic stack manipulation words ----------
;
; drop ( a -- )         Remove the top item on the data stack

code_drop:
    pop ebx             ; ebx (top of stack cache) it taken from the item
    next                ; below, and the stack pointer is updated

; dup ( a -- a a )      Create a duplicate copy of the top data stack item

code_dup:
    push ebx            ; Preserve the old top item (ebx), while creating a
    next                ; new second item equal to that one

; swap ( a b -- b a )   Swap the top two stack items

code_swap:
    xchg ebx,[esp]      ; Exchange the items in ebx (top of data stack) and
    next                ; [esp] (second item of data stack)

; nip ( a b -- b )      Drop the second stack item (equivalent to "swap drop")

code_nip:
    add esp,4           ; Drop the second stack item
    next

; r> ( -- a )           Fetch one value from the return stack

code_rto:
    push ebx            ; Save old TOS
    mov ebx,[ebp]       ; Get new value from return stack
    add ebp,4
    next

; >r ( a -- )           Push one value to the return stack

code_tor:
    sub ebp,4
    mov [ebp],ebx       ; Push value to return stack
    pop ebx             ; Drop it from data stack
    next

; ---------- Arithmetics words ----------
;
; + ( a b -- c )        c = a + b

code_add:
    add [esp],ebx       ; Add top item to second item
    pop ebx             ; Make the second item the new top item
    next

; - ( a b -- c )        c = a - b

code_sub:
    sub [esp],ebx       ; Subtract top item from second item
    pop ebx             ; Make the second item the new top item
    next

; and ( a b -- c )      c = a & b  (bitwise AND)

code_and:
    and [esp],ebx       ; AND second item with top item
    pop ebx             ; Make the second item the new top item
    next

; or ( a b -- c )       c = a | b  (bitwise OR)

code_or:
    or [esp],ebx        ; OR second item with top item
    pop ebx             ; Make the second item the new top item
    next

; ---------- Memory access words ----------
;
; @ ( a -- x )          Fetch word at address a

code_fetch:
    mov ebx,[ebx]       ; ebx was the address, becomes the word
    next

; ! ( x a -- )          Store word x to address a

code_store:
    pop eax             ; eax = word to store, ebx = address
    mov [ebx],eax
    pop ebx
    next

; ---------- Program flow control words ----------
;
; branch ( -- )         Read the address following the branch instruction
;                       i a high-level word, and jump there

code_branch:
    mov esi,[esi]       ; Read the new address into the interpretation pointer
    next

; ?nbranch ( a -- )     Like branch, but only if the top of the stack is zero

code_cnbranch:
    test ebx,ebx        ; Check if top of data stack is 0
    pop ebx             ; Remove the top item
    jz code_branch      ; Jump to the branch word if it turned out to be 0
    add esi,4           ; Otherwise we'll need to update the interpretation
    next                ; pointer to skip the branch address

; 0= ( a -- b )         Logical negation (equivalent to "0 =")

code_zeq:
    test ebx,ebx        ; Is the top of the data stack zero?
    jz .zero            ; Yes, so set it to -1 (true)
    mov ebx,0           ; No, so set it to 0 (false)
    next
.zero:
    mov ebx,-1
    next

; = ( a b -- c )        Return -1 if a is equal to b, 0 otherwise

code_eq:
    pop eax             ; Get a
    cmp eax,ebx         ; Compare to b
    jz .equal           ; Are they equal?
    mov ebx,0           ; No, so return 0 (false)
    next
.equal:
    mov ebx,-1          ; Yes, so return -1 (true)
    next

; ---------- I/O words ----------
;
; type ( s -- )         Print the zero-terminated string to stdout

code_type:
    movzx eax,byte[ebx] ; Load one byte and increase pointer (movzx fills the
    inc ebx             ; upper bits of eax with zero bits)
    test eax,eax        ; End of string?
    jz .done
    push eax            ; No, call putchar(character)
    call putchar
    add esp,4
    jmp code_type       ; Print next character
.done:
    next

; . ( a -- )            Print the number a

code_dot:
    push ebx            ; Push the number
    push dword str_dotformat
    call printf         ; printf("%d",a)
    add esp,8
    pop ebx             ; Drop the number
    next

; emit ( a -- )         Print the character a

code_emit:
    push ebx            ; Push the character
    call putchar        ; putchar(a)
    add esp,4
    pop ebx             ; Drop the character
    next

; key ( -- a )          Read a character from stdin

code_key:
    call getchar        ; Read the character to eax
    push ebx
    mov ebx,eax         ; Push it to the data stack
    next

; ms ( a -- )           Wait for a milliseconds

code_ms:
    mov eax,1000
    mul ebx             ; Multiply a by 1000
    push eax
    call usleep         ; usleep(a*1000)
    add esp,4
    pop ebx
    next

; ---------- Compiler related words ----------
;
; lookup ( s -- a|0 )   Find the word with the name pointed to by "s", and
;                       return the address of its header (or 0, if no such
;                       word was found).

code_lookup:
    push edi
    mov edi,[last]          ; edi will always point to the header of a word
                            ; we're going to compare the name string to
.try_next:
    test edi,edi            ; Are all bits in edi 0? Then we've reached the
    jz .done                ; end of the header list, and we failed
    lea eax,[edi+0x0c]      ; eax points to the name of the current word to be
                            ; compared to the name string
    push eax
    push ebx
    call strcmp             ; eax = strcmp(name, name of current header)
    add esp,8

    test eax,eax            ; Is eax 0? Then the strings match, and we return
    jz .done                ; this header

    mov edi,[edi+4]         ; Make edi point to the next header in the list
    jmp .try_next           ; And check this one against the given name
.done:
    mov ebx,edi             ; Top of stack points to header adddress (or 0 in
    pop edi                 ; case no word was found)
    next

; read ( s -- )             Read a word from stdin to the specified address
code_read:
    call getchar            ; al = getchar()
    push eax
    call putchar
    pop eax
    cmp al,0x20             ; Is al a special ASCII character (0x20 and
    jbe code_read           ; below)? Then it's most likely a blankspace and
                            ; we will ignore it
.read_char:                 ; Inside this loop we're reading a word until a
                            ; blankspace is found
    mov [ebx],al            ; We want to write the character we just read to
    inc ebx                 ; the string buffer given, and update the pointer
    call getchar            ; Read another character
    push eax
    call putchar
    pop eax
    cmp al,0x20             ; Did we find the ending whitespace?
    ja .read_char           ; No, so read and copy another character

    mov byte[ebx],0         ; End of word, so make the buffer zero-terminated
    pop ebx                 ; Drop buffer address
    next

; mode ( -- a )             Return the current mode (0 = compilation, 1 =
;                           interpretation)

code_mode:
    push ebx                ; Save old top of data stack
    mov ebx,[mode]          ; Replace it with the current mode
    next

; last ( -- a )             Get the address of the pointer to the last header

code_last:
    push ebx                ; Save old top of data stack
    mov ebx,last            ; Replace it with the address of this pointer
    next

; herecode ( -- a )         Get the current top of the code heap

code_herecode:
    push ebx                ; Save old top of data stack
    mov ebx,[codeheap_ptr]  ; Set it to the address wanted
    next

; ,code ( a -- )            Write a word to the code heap, and update the code
;                           heap pointer

code_commacode:
    mov eax,[codeheap_ptr]  ; eax points to the heap location we'll write to
    add dword[codeheap_ptr],4   ; update heap pointer
    mov [eax],ebx           ; Store the top data stack item
    pop ebx                 ; ..and get rid of it afterwards
    next

; c,code ( a -- )           Write a byte to the code heap, and update the code
;                           heap pointer

code_ccommacode:
    mov eax,[codeheap_ptr]  ; eax points to the heap location we'll write to
    inc dword[codeheap_ptr] ; update heap pointer
    mov [eax],bl            ; Store the top data stack item
    pop ebx                 ; ..and get rid of it afterwards
    next

; , ( a -- )                Write a word to the data heap

code_comma:
    mov eax,[dataheap_ptr]  ; eax points to the heap location we'll write to
    add dword[dataheap_ptr],4   ; update heap pointer
    mov [eax],ebx           ; Store the top data stack item
    pop ebx                 ; ..and get rid of it afterwards
    next

; s>i ( s -- x n )          Convert the string to an integer. Returns the
;                           integer in x (0 for an invalid number), and a flag
;                           in n (0 for invalid number, -1 for valid number)

code_stoi:
    push dword 0            ; base 0 (meaning default base is 10, but others
                            ; are accepted, for example with a 0x prefix)
    push dword first_error  ; Will be pointing to 0 if the number was valid
    push ebx                ; The address of the string
    call strtol
    add esp,12

    push eax                ; Second stack item = the number itself
    mov eax,[first_error]
    cmp byte[eax],0         ; Was the number valid?
    jz .valid
    mov ebx,0               ; No, return false
    next
.valid:
    mov ebx,-1              ; Yes, return true
    next

; [                         Switch to interpretation mode

code_imode:
    mov dword[mode],mode_interpret
    next

; ]                         Switch to compilation mode

code_cmode:
    mov dword[mode],mode_compile
    next

; process ( s -- )          Interpret or compile the word whose name is in 's'
;                           If no such word is found, and the name is a valid
;                           number, that number will either be pushed to the
;                           data stack (interpretation mode) or be compiled as
;                           a "lit" call (compilation mode).

code_process:
    call nest               ; This is a high-level word
    dd _dup,_lookup         ; Save the string address and try to find a word
                            ; with that name
    dd _dup                 ; Save result (0 = not found, or header address)
    dd _cnbranch,.noword    ; Branch if no word was found
    dd _nip                 ; Drop the string - we won't need it again here
    dd _dup,_lit,8,_add     ; Save header address, and add 8 to it
    dd _fetch               ; Fetch the word there (flag word)
    dd _lit,1,_and,_zeq     ; AND with 1 (to check immediate flag)
    dd _cnbranch,.interpretw; If it's set, interpret this word. Otherwise:
    dd _mode                ; Get mode (0 = compile, 1 = interpret)
    dd _cnbranch,.compilew  ; Choose if we should compile or interpret word
.interpretw:
    dd _execute,_exit       ; Execute the word and return
.compilew:
    dd _commacode,_exit     ; Compile the word and return
.noword:
    dd _drop                ; Drop the 0 (not found) from "lookup"
    dd _stoi                ; Try to convert the string to an integer instead
    dd _cnbranch,.fail      ; If that didn't work, print an error message
    dd _mode,_zeq           ; Are we compiling or interpreting?
    dd _cnbranch,.done      ; Interpreting, just leave the number and return
                            ; Compiling, so do this:
    dd _lit,_lit            ; Push the address of the "lit" word header
    dd _commacode           ; Write it to the word we are compiling
    dd _commacode           ; After it, write the integer (so that it will be
                            ; pushed when the word we are compiling is run)
.done:
    dd _exit                ; Return
.fail:
    dd _drop                ; Drop the 0 from "s>i"
    dd _lit,str_notfound    ; Push the address of the "word not found" string
    dd _type                ; Print it
    dd _exit                ; Return

; header ( s -- )           Create a word header with the given name

code_header:
    mov ecx,[dataheap_ptr]  ; ecx points to the end of the data heap, which is
                            ; where the header will go
    mov dword[ecx],code_nop ; By default, make the header link to a dummy word
    mov eax,[last]          ; eax points to the previous header created
    mov [ecx+4],eax         ; Link this new header to it
    mov [last],ecx          ; And make the new header the start of the list
    mov dword[ecx+8],flag_none ; Initially set no flags
    add ecx,0x0c            ; ecx now points to where we want the name string
.copyname:
    mov al,[ebx]            ; Load a byte and update the string pointer
    inc ebx
    mov [ecx],al            ; Store the byte
    inc ecx
    cmp al,0                ; Zero byte (end of string)?
    jnz .copyname           ; No, copy another byte
    add ecx,3               ; Make ecx (the new end of the heap) aligned to a 
    and cl,0xfc             ; four-byte boundary. Bit magic.
    mov [dataheap_ptr],ecx  ; Update the data heap pointer to the end of this
    pop ebx                 ; header
    next

; : wordname                Define a new word with the given word name

code_colon:
    call nest               ; This is a high-level word
    dd _lit,wordbuf,_dup
    dd _read                ; Read a word to the buffer
    dd _header              ; Create a header with that name
    dd _herecode            ; Get address of code heap
    dd _last,_fetch         ; Get address of the code pointer in the new header
    dd _store               ; Store the code heap address there
    dd _lit,0xe8            ; "call" instruction opcode
    dd _ccommacode          ; Store it to the code heap
    dd _lit,nest            ; Get address of the "nest" word
    dd _herecode,_sub       ; Subtract the address of the op code
    dd _lit,4,_sub          ; Add 4 to get the correct relative address
    dd _commacode           ; Store this new address
    dd _cmode               ; Switch to compilation mode
    dd _exit                ; Return

; ;                         Semicolon - end word definition.

code_semicolon:
    call nest               ; High-level word
    dd _lit,_exit           ; Push address of the header of the exit word
    dd _commacode           ; Compile it to the current word
    dd _imode               ; Switch to interpretation mode
    dd _exit

; wordbuf ( -- a )          Push the address of the word buffer.

code_wordbuf:
    call nest
    dd _lit,wordbuf,_exit

; param ( -- a )            Get a "parameter" to this word from the thread.

code_param:
    push ebx                ; Save old TOS cache
    mov eax,[ebp]
    add dword[ebp],4
    mov ebx,[eax]
    next

; ---------- System interfacing words ----------

code_bye:
    push dword 0
    call exit

code_quit:
    call nest           ; This is a high-level word
_quit_body:
    dd _lit,wordbuf     ; Push address of word input buffer
    dd _read            ; Read it
    dd _lit,wordbuf     ; Push address of word input buffer
    dd _process         ; Interpret/compile it
    dd _branch,_quit_body

; Initialize the Forth system, by doing the following:
;  (1) set ebp, the interpreter's return stack pointer
;  (2) set esi, the interpreter's instruction pointer, to execute the word
;      "quit" (equivalent to C's "main()" function, but Forth tradition calls
;      this the "quit" loop. This word should never return, but call exit() to
;      terminate the process whenever the user wants to quit
;  (3) execute the code in the "next" macro to start the interpreter

initialize:
    mov ebx,0           ; Initialize top of data stack to 0, for no particular
                        ; reason
    mov ebp,returnstack_top
    mov esi,_quit_body  ; Note that esi must point to the address of the
    next                ; header of the first word in "quit"

; int main()
;
; The main function that libc expects to find. We will never return from this,
; its only purpose is to initialize the interpreter.
main:
    jmp initialize

; End of code section


; Section for initialized data, such as the lits of built-in words, and some
; global variables.
section '.data' writeable

; (some) inner interpreter words
_exit:      word code_exit,     'exit',     flag_none
_lit:       word code_lit,      'lit',      flag_none
_execute:   word code_execute,  'execute',  flag_none

; Stack manipulation words
_drop:      word code_drop,     'drop',     flag_none
_dup:       word code_dup,      'dup',      flag_none
_swap:      word code_swap,     'swap',     flag_none
_nip:       word code_nip,      'nip',      flag_none
_rto:       word code_rto,      'r>',       flag_none
_tor:       word code_tor,      '>r',       flag_none

; Arithmetics words
_add:       word code_add,      '+',        flag_none
_sub:       word code_sub,      '-',        flag_none
_and:       word code_and,      'and',      flag_none
_or:        word code_or,       'or',       flag_none

; Memory access words
_fetch:     word code_fetch,    '@',        flag_none
_store:     word code_store,    '!',        flag_none

; Program flow control words
_branch:    word code_branch,   'branch',   flag_none
_cnbranch:  word code_cnbranch, '?nbranch', flag_none
_zeq:       word code_zeq,      '0=',       flag_none
_eq:        word code_eq,       '=',        flag_none

; I/O words
_type:      word code_type,     'type',     flag_none
_dot:       word code_dot,      '.',        flag_none
_emit:      word code_emit,     'emit',     flag_none
_key:       word code_key,      'key',      flag_none
_ms:        word code_ms,       'ms',       flag_none

; Compiler related words
_mode:      word code_mode,     'mode',     flag_none
_last:      word code_last,     'last',     flag_none
_read:      word code_read,     'read',     flag_none
_lookup:    word code_lookup,   'lookup',   flag_none
_header:    word code_header,   'header',   flag_none
_stoi:      word code_stoi,     's>i',      flag_none
_process:   word code_process,  'process',  flag_none
_comma:     word code_comma,    ',',        flag_none
_imode:     word code_imode,    '[',        flag_immediate
_cmode:     word code_cmode,    ']',        flag_none
_herecode:  word code_herecode, 'herecode', flag_none
_commacode: word code_commacode,',code',    flag_none
_ccommacode: word code_ccommacode,'c,code', flag_none
_colon:     word code_colon,    ':',        flag_none
_semicolon: word code_semicolon,';',        flag_immediate
_wordbuf:   word code_wordbuf,  'wordbuf',  flag_none
_param:     word code_param,    'param',    flag_none

; System interfacing words
_bye:       word code_bye,      'bye',      flag_none

; Global variables
dataheap_ptr    dd  dataheap_start      ; Current top of the data heap
codeheap_ptr    dd  codeheap_start      ; Current top of the code heap
last            dd  previous_word       ; Current start of the header list
mode            dd  mode_interpret      ; Start in interpretation mode

; String contstants
str_notfound    db  'Word not found',0x0a,0
str_dotformat   db  '%d',0              ; Format string used for the "." word

; End of initialized data section

; Uninitialized data. This will be used for the data heap, code heap, and
; return stack
section '.bss' executable writeable align 0x1000

; Sizes of the data areas
;
; Note: make these page aligned (4kB) to avoid flushing of the code cache when
; data is written to the surrounding data heap and return stack areas

returnstack_size    = 0x1000    ; 4kB return stack
dataheap_size       = 0x10000   ; 64kB of data
codeheap_size       = 0x10000   ; 64kB of code
                                ; (it's safe to make these much larger, all
                                ;  memory specified here will not be allocated
                                ;  until it's needed)

returnstack_start:
    rb  returnstack_size        ; rb = reserve byte(s)
returnstack_top:                ; This is where the top of the return stack
                                ; is located, since we're filling it from the
                                ; top down we want to know this address

dataheap_start:
    rb  dataheap_size

codeheap_start:
    rb  codeheap_size

wordbuf:                        ; Buffer to store the current word name
    rb  0x40

first_error:                    ; strtol() error pointer
    rd  1

